from typing import Optional 
import spacy
import codecs
from nltk.tokenize.treebank import TreebankWordDetokenizer
from collections import defaultdict
from qasem import data_dir

connective_file_default_path = data_dir / "connectives_small_set.txt"

def get_connectives(conn_file):
    conns = defaultdict(lambda : '')
    infile = codecs.open(conn_file, 'r')
    for line in infile.readlines():
        line = line.split(',')
        conns[line[0]] = 'ok'
    return conns

def splitter(splittoken, mylist, post = True, add_white_sp = False):
    splitlist = []
    for span in mylist:
        if splittoken in span:
            newsplit = span.split(splittoken)
            newsplitlist = []
            if add_white_sp:
                if post:
                    prev_item_emtpy = False
                    for count, item in enumerate(newsplit):
                        if item.strip() == '':
                            prev_item_emtpy = True
                        elif prev_item_emtpy:
                            newsplitlist.append(item.strip() + ' ' + splittoken.strip())
                            prev_item_emtpy = False
                        elif count == len(newsplit)-1:
                            newsplitlist.append(item.strip())
                        else:
                            prev_item_emtpy = False
                            newsplitlist.append(item.strip()+' '+splittoken.strip())
                else:
                    for count, item in enumerate(newsplit):
                        if count == 0:
                            newsplitlist.append(item.strip())
                        else:
                            newsplitlist.append(splittoken.strip()+' '+item.strip())
            else:
                if post:
                    prev_item_emtpy = False
                    for count, item in enumerate(newsplit):
                        if item.strip() == '':
                            prev_item_emtpy = True
                        elif prev_item_emtpy:
                            newsplitlist.append(item.strip() + splittoken.strip())
                            prev_item_emtpy = False
                        elif count == len(newsplit) - 1:
                            newsplitlist.append(item.strip())
                        else:
                            prev_item_emtpy = False
                            newsplitlist.append(item.strip()  + splittoken.strip())
                else:
                    for count, item in enumerate(newsplit):
                        if count == 0:
                            newsplitlist.append(item.strip())
                        else:
                            newsplitlist.append(splittoken.strip() + item.strip())
            splitlist += newsplitlist
        else:
            splitlist.append(span)
    return splitlist
    
    
    
    
class CandidateFinder():

    def __init__(self, conn_file: Optional[str] = None):
        conn_file = conn_file or connective_file_default_path
        self.connectives = get_connectives(conn_file)
        self.detokenizer = TreebankWordDetokenizer()
        self.nlp = spacy.load("en_core_web_sm")
        self.covered = defaultdict(lambda : '')
        
        
    def num_candidates(self, sentence):
        sentence = sentence.strip().split('\t')
        text = sentence[0]
        text = text.replace(' -LRB- ', ' ')
        text = text.replace(' -RRB- ', ' ')
        text = text.replace(' -LSB- ', ' ')
        text = text.replace(' -RSB- ', ' ')
        text = self.detokenizer.detokenize(text.split(' '))
        text = text.strip()
        text = text.replace('"', ' "" ')
        text = text.replace("''", ' "" ')
        text = text.replace("``", ' "" ')
        text = text.replace("'", "\\'")
        text = text.replace(" , ", ", ")
        # if text not in self.covered:
        #    self.covered[text] = True
        punc_split = splitter(', ', [text])
        punc_split = splitter('," ', punc_split)
        punc_split = splitter('; ', punc_split)
        punc_split = splitter(': ', punc_split)
        for conn in self.connectives.keys():
            if ' '+conn+' ' in text:
                punc_split = splitter(' '+conn+' ' , punc_split, post=False, add_white_sp = True)
        doc = self.nlp(text)
        word_to_pos = defaultdict(lambda : '')
        for count, token in enumerate(doc):
            word_to_pos[token.text] = token.tag_
        ids = []
        # Split if more than 1 verb:
        # VBD / VBN / VBP
        for spannum, span in enumerate(punc_split):
            verb_collector = []
            inner_verbs = []
            verb_counter = 0
            noun_collector = []
            other_collector = []
            prev_word_verb = False
            for wordnum, word in enumerate(span.split(' ')):
                word = word.strip()
                id = 0
                for i in range(spannum):
                    id += len(punc_split[i].split(' '))
                id += wordnum
                if word_to_pos[word] == '' and len(word) > 0:
                    if word[-2:].strip() in ['."', ',"']:
                        word = word[:-2]
                    elif word[-1] in [',', ';', ':', '.', '"']:
                        word = word[:-1]
                    elif word[-2:] in ["'d", "'s"]:
                        word = word[-2:]
                    elif word[-3:] in ["'re"]:
                        word = word[-3:]
                    elif word[-3:] in ["n't"] and word != "won't":
                        word = word[:-3]
                if word_to_pos[word].startswith('N'):
                    noun_collector.append(id)
                    prev_word_verb = False
                elif word_to_pos[word].startswith('RB'):
                    prev_word_verb = True
                elif word_to_pos[word].startswith('V'):
                    if word in ['said', 'according', 'spoke']:
                        pass
                    elif not prev_word_verb:
                        if len(inner_verbs)>0:
                            verb_collector.append(inner_verbs)
                        inner_verbs = []
                        inner_verbs.append(id)
                    else:
                        inner_verbs.append(id)
                    verb_counter += 1
                    prev_word_verb = True
                elif word == "n't":
                    prev_word_verb = True
                else:
                    other_collector.append(id)
                    prev_word_verb = False
            verb_collector.append(inner_verbs)
            if verb_counter == 0 and len(noun_collector)>0:
                if span.split(' ')[0].lower() in self.connectives and span.split(' ')[0].lower() != 'and':
                    ids.append(noun_collector[0])
            elif verb_counter == 0 and len(noun_collector) == 0:
                if span.split(' ')[0].lower() in self.connectives:
                    try:
                        ids.append(other_collector[1])
                    except IndexError:
                        pass
            elif verb_counter > 0:
                for v in verb_collector:
                    if len(v)>0:
                        ids.append(v[-1])
        return len(ids)

        
        


from typing import Tuple, Iterable, List, Any, Dict
import json
import itertools 

OIE = Tuple[str, ...]
default_layers_included = ["qasrl"]

class OpenIEConverter:
    """
    Converts a set of QA-SRL QAs into Open Information Extraction tuples.
    The OpenIE tuples (=propositions) have >3 slots, where first is subject, 
    second is predicate, and the rest are further arguments. 
    """
    def __init__(self, layers_included: List[str] = default_layers_included):
        self.layers_included = layers_included
    
    def convert_qadiscourse_qas(self, qas) -> List[OIE]:
        """ Converts QADiscourse information into OpenIE propositions. """
        # Future: consider how to convert qadiscourse QAs into propositions
        # qas = self.filter_qadiscourse_redundant_QAs(qas)
        # return [proposition 
        #         for qa in qas 
        #         for proposition in self.convert_single_qadiscourse_qa(qa)]
        
        # Currently, don't translate QADiscourse QAs into propositions
        return []
        
    def convert_single_qadiscourse_qa(self, qa) -> OIE:
        """ Converts a single QADiscourse qa into a single OpenIE propositions. """
        raise NotImplementedError()
    
    def filter_qadiscourse_redundant_QAs(self, qas) -> List[Dict[str, str]]:
        # remove equivalent QAs
        return [dict(t) for t in {tuple(qa.items()) for qa in qas}]
    
    
    def convert_single_qasrl_predicate(self, pred_info, sentence) -> List[OIE]:
        """ Converts QASRL information of a verb or a nominalization into OpenIE propositions. """
        # collect arguments (grouped by question) along with their index
        args = [[(ans, sentence.index(ans)) for ans in qa["answers"]]
                 for qa in pred_info["QAs"]]
        # sort by first occurrence
        args = list(sorted(args, key=lambda answers: min(t[1] for t in answers)))
        # omit repetitions (full or partial)
        def get_answers_without_repetitions(answers: List[Tuple[str, int]]):
            answers_no_repetitions = answers.copy()
            i=1
            while i < len(answers_no_repetitions): 
                if answers_no_repetitions[i] in answers_no_repetitions[i-1]:
                    del answers_no_repetitions[i]
                else:
                    i += 1
            return answers_no_repetitions
        args = [get_answers_without_repetitions(answers) for answers in args]

            
        # omit indices
        args_strs = [[a[0] for a in arg] for arg in args]
        # add predicate to args
        # args.insert(1, [(pred_info["predicate"], pred_info["predicate_idx"])])
        args_strs.insert(1, [pred_info["predicate"]])
        # yield Cartesian product of args
        propositions = []
        for tup in itertools.product(*args_strs):
            propositions.append(tup)
        return propositions
        
    
    def convert_single_sentence(self, sent_info, sentence: str) -> List[OIE]:
        propositions: List[OIE] = []
        # Collect from QASRL and QANom layers
        qasrl_propositions: List[OIE] = []
        qadisc_propositions: List[OIE] = []
        for layer in self.layers_included:
            if layer in ("qasrl", "qanom"):
                # iterate verbal/nominal predicates
                for pred_info in sent_info[layer]:
                    qasrl_propositions.extend(self.convert_single_qasrl_predicate(pred_info, sentence))
            if layer == "qadiscourse":
                qadisc_propositions.extend(self.convert_qadiscourse_qas(sent_info["qadiscourse"]))
        propositions = qasrl_propositions + qadisc_propositions
        return propositions
                
    
def test_openie_converter():
    converter = OpenIEConverter()
    input_sentences = ["The doctor was very interested in Luke 's treatment as he was not feeling well .",
                       "Tom brings the dog to the park."]
    pipe_output = json.load(open("tmp/example_output.json"))
    oie_outputs = [converter.convert_single_sentence(sent_info, sent)
                   for sent_info, sent in zip(pipe_output, input_sentences)]
    print(oie_outputs)
    

if __name__ == "__main__":
    test_openie_converter()


from qasem.end_to_end_pipeline import QASemEndToEndPipeline
import spacy
from pprint import pprint
from tqdm import tqdm

if __name__ == "__main__":
    layers = [
        "qasrl",
        "qanom",
        "qadiscourse",
    ]
    pipe = QASemEndToEndPipeline(layers, device=-1, nominalization_detection_threshold=0.75, contextualize = True)
    sentences = ["The doctor was very interested in Luke 's treatment as he was not feeling well .", "Tom brings the dog to the park.", "I hate cats."]
    outputs = pipe(sentences*30, verbose=True)
    pprint(outputs)
    # print()
    # sentences = ["the ball is red ."]
    # sentences = ["the construction of the officer 's building finished right after the beginning of the destruction of the previous construction ."]
    # sentences = ["The Veterinary student was interested in Luke 's treatment of sea animals .", "John ate an apple because he was hungry ."]
    # outputs = pipe(sentences)
    # pprint(outputs)

    # pipe = QASemEndToEndPipeline(['qasrl'], nominalization_detection_threshold=0.75, contextualize = False)
    # sentences = ["The doctor was interested in Luke 's treatment .", "The construction of the officer 's building finished right after the beginning of the destruction of the previous construction ." , "The Veterinary student was interested in Luke 's treatment of sea animals .", "Tom brings the dog to the park."]
    # outputs = pipe(sentences)
    # print(outputs)
    # print()
    #
    # pipe = QASemEndToEndPipeline(['qadiscourse'])
    # sentence = "The construction of the officer 's building will finish after the rain would stop ."
    # print(pipe([sentence]))
    # print()
    # # print()
    #
    #
    # # print('\n')
    # res3 = pipe(['Tom brings the dog to the park.'])
    # print(res3)
    # print()
    #
    #
    # res3 = pipe(['.', 'Tom brings the dog.' ])
    # print(res3)




from pathlib import Path
package_homedir = Path(__file__).parent
data_dir = package_homedir / "data"
# version
with open(package_homedir / "version.txt", "r") as f:
    __version__ = f.read().strip()
